#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/config.h>
#include <asm/brecis/prom.h>
#include <asm/brecis/BrecisSysRegs.h>

extern unsigned char V_fsize;

static __u8 b_flash_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

static __u16 b_flash_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

static __u32 b_flash_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

static void b_flash_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

static void b_flash_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

static void b_flash_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

static void b_flash_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

static void b_flash_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

/* board and partition description */
static struct mtd_partition ax112w_partitions_4M[] = 
{
    /* The first stage bootloader */
    {
	"Bootldr",
	56 * 1024,
	0,
    },    
    {
	"Syscfg",
	8 * 1024,
	56 * 1024,
    },
    /* The second stage bootloader */
    {
	"Core",
	0x380000-(64*1024), // 3.5 MB - 64K
	64 * 1024,
    },
    {
	"Config",
	0x80000, // 0.5 MB
	0x380000, 
    }
};

#define NUM_PARTITIONS_4M ( sizeof( ax112w_partitions_4M) / sizeof( struct mtd_partition ) )

/* board and partition description */
static struct mtd_partition ax112w_partitions_8M[] = 
{
    /* The first stage bootloader */
    {
	"Bootldr + syscfg",
	64 * 1024,
	0x400000,
    },    
    {
	"Dummy 64K",
	64 * 1024,
	0,
    },
    /* The second stage bootloader */
    {
	"Core",
	0x380000-(64*1024), // 3.5 MB - 64K
	(0x400000+(64 * 1024)),
    },
    {
	"Config",
	0x80000, // 0.5 MB
	(0x400000+(0x380000)), 
    },
    {
	"Extra",
	0x400000-(64*1024), // 4MB - 64K
	64*1024, 
    }
};

#define NUM_PARTITIONS_8M ( sizeof( ax112w_partitions_8M) / sizeof( struct mtd_partition ) )


static struct map_info v_flash_map = {
	name: "AX-112W flash",
	size: 0,
	buswidth: 0,
	read8: b_flash_read8,
	read16: b_flash_read16,
	read32: b_flash_read32,
	copy_from: b_flash_copy_from,
	write8: b_flash_write8,
	write16: b_flash_write16,
	write32: b_flash_write32,
	copy_to: b_flash_copy_to,
};

int __init init_ax112w_flash(void)
{
	struct mtd_info *mymtd;

	static char	name[] = "flash0";
	unsigned int	addr, size;
	int	buswidth = 1;
	int	result;
	char	*env;

	env = prom_getenv(name);
	buswidth=1;
	if (!env) {
		addr=0xbfc00000;
		size=0x400000;
	} else {
		result = sscanf(env, "%x:%x", &addr, &size);
		if (result < 2) {
			addr=0xbfc00000;
			size=0x400000;
		}
	}

	printk(KERN_NOTICE "flash device: [%x], size: %x\n", addr, size);

	v_flash_map.map_priv_1 = (unsigned long) ioremap(addr, size);
	if (!v_flash_map.map_priv_1) {
		printk(KERN_WARNING "Failed to ioremap\n");
		return -EIO;
        }
	v_flash_map.size = size;
	v_flash_map.buswidth = buswidth;

	mymtd = (struct mtd_info *) do_map_probe("cfi_probe", &v_flash_map);
	if (mymtd) {
		mymtd->module = THIS_MODULE;
		v_flash_map.map_priv_2 = (unsigned long)mymtd;
		if (size==0x800000) {
			V_fsize=8;
			add_mtd_partitions(mymtd, ax112w_partitions_8M, NUM_PARTITIONS_8M);
		} else {
			V_fsize=4;
			add_mtd_partitions(mymtd, ax112w_partitions_4M, NUM_PARTITIONS_4M);
		}
	} else
		return -ENXIO;
	return 0;
}

static void __exit cleanup_ax112w_flash(void)
{
        struct mtd_info *mymtd;

        mymtd = (struct mtd_info *)v_flash_map.map_priv_2;
	if (mymtd) {
		del_mtd_partitions(mymtd);
		map_destroy(mymtd);
	}
	if (v_flash_map.map_priv_1) {
		iounmap((void *)v_flash_map.map_priv_1);
		v_flash_map.map_priv_1 = 0;
	}
}

module_init(init_ax112w_flash);
module_exit(cleanup_ax112w_flash);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("ASUS");
MODULE_DESCRIPTION("AX112W");
