///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2003 C++ to C Made by PC32 Lsshi
// All rights reserved. 

///////////////////////////////////////////////////////////////////////////

#ifndef PORT_LINKED_LIST_H
#define PORT_LINKED_LIST_H

#ifndef NULL
#define NULL 0
#endif

#define LIST_SUCCESS 1
#define LIST_FAIL 0




/****************************************************************************
 * Name: free_routine
 *
 *  Description:
 *     Function for freeing list items
 *****************************************************************************/
typedef void (*port_free_function)(void *arg);


/****************************************************************************
 * Name: ListNode
 *
 *  Description:
 *      linked list node. stores generic item and pointers to next and prev.
 *      Internal Use Only.
 *****************************************************************************/
typedef struct PORTLISTNODE
{
  struct PORTLISTNODE *next; //next node
  void *item; //item
} Port_ListNode;

/****************************************************************************
 * Name: LinkedList
 *
 *  Description:
 *      linked list (no protection). Internal Use Only.
 *      Because this is for internal use, parameters are NOT checked for 
 *      validity.
 *      The first item of the list is stored at node: head->next
 *      The last item of the list is stored at node: tail->prev
 *      If head->next=tail, then list is empty.
 *      To iterate through the list:
 *
 *       LinkedList g;
 *       ListNode *temp = NULL;
 *       for (temp = ListHead(g);temp!=NULL;temp = ListNext(g,temp))
 *       {
 *        }
 *
 *****************************************************************************/
typedef struct PORTLINKEDLIST
{
  Port_ListNode* head; //head, first item
  Port_ListNode* tail; //head, last item
  
  long size;      //size of list
  port_free_function free_func; //free function to use
  
} PortLinkedList;

/****************************************************************************
 * Function: PortListInit
 *
 *  Description:
 *      Initializes LinkedList. Must be called first.
 *      And only once for List.
 *  Parameters:
 *      list  - must be valid, non null, pointer to a linked list.
 *      free_func - function used to free items. (May be NULL)
 *  Returns:
 *      0 on success, EOUTOFMEM on failure.
 *****************************************************************************/
int PortListInit(PortLinkedList *list, port_free_function free_func);


/****************************************************************************
 * Function: PortListAddTail
 *
 *  Description:
 *      Adds a node to the tail of the list.
 *      Node gets added immediately before list.tail.
 *  Parameters:
 *      LinkedList *list  - must be valid, non null, pointer to a linked list.
 *      void * item - item to be added
 *  Returns:
 *      The pointer to the ListNode on success, NULL on failure.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
Port_ListNode *PortListAddTail(PortLinkedList *list, void *item);


/****************************************************************************
 * Function: ListDelNode
 *
 *  Description:
 *      Removes a node from the list
 *      The memory for the node is freed.
 *  Parameters:
 *      LinkedList *list  - must be valid, non null, pointer to a linked list.
 *      ListNode *dnode - done to delete.
 *      freeItem - if !0 then item is freed using free function.
 *                 if 0 (or free function is NULL) then item is not freed
 *  Returns:
 *      The pointer to the item stored in the node or NULL if the item is freed.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
int PortListDelNode(PortLinkedList *list,Port_ListNode *dnode, int freeItem);

/****************************************************************************
 * Function: ListDestroy
 *
 *  Description:
 *      Removes all memory associated with list nodes. 
 *      Does not free LinkedList *list. 
 *    
 *  Parameters:
 *      LinkedList *list  - must be valid, non null, pointer to a linked list.
 *      freeItem - if !0 then items are freed using the free_function.
 *                 if 0 (or free function is NULL) then items are not freed.
 *  Returns:
 *      0 on success. Always returns 0.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
int PortListDestroy(PortLinkedList *list, int freeItem);


/****************************************************************************
 * Function: ListHead
 *
 *  Description:
 *      Returns the head of the list.
 *    
 *  Parameters:
 *      LinkedList *list  - must be valid, non null, pointer to a linked list.
 *  
 *  Returns:
 *      The head of the list. NULL if list is empty.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
Port_ListNode* PortListHead(PortLinkedList *list);


/****************************************************************************
 * Function: ListSize
 *
 *  Description:
 *     Returns the size of the list.
 *  Parameters:
 *      LinkedList *list  - must be valid, non null, pointer to a linked list.
 
 *  Returns:
 *      The number of items in the list.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
int PortListSize(PortLinkedList* list);


#endif //LINKED_LIST_H
