#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/sysmacros.h>

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <time.h>
#include <memory.h>
#include <utmp.h>
#include <mntent.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <termios.h>
#include <unistd.h>

#if __GLIBC__ >= 2
#include <asm/types.h>		/* glibc 2 conflicts with linux/types.h */
#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <netinet/if_ether.h>
#else
#include <linux/types.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/route.h>
#include <linux/if_ether.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>


#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

#define SIN_ADDR(x)	(((struct sockaddr_in *) (&(x)))->sin_addr.s_addr)
#define SET_SA_FAMILY(addr, family)			\
	memset ((char *) &(addr), '\0', sizeof(addr));	\
	addr.sa_family = (family);

/*
 * /proc/net/route parsing stuff.
 */
#define ROUTE_MAX_COLS	12
FILE *route_fd = (FILE *) 0;
static char route_buffer[512];
static int route_dev_col, route_dest_col, route_gw_col;
static int route_flags_col, route_mask_col;
static int route_num_cols;

static int open_route_table (void);
static void close_route_table (void);
static int read_route_table (struct rtentry *rt);

/********************************************************************
 *
 * close_route_table - close the interface to the route table
 */

static void close_route_table (void)
{
    if (route_fd != (FILE *) 0) {
	fclose (route_fd);
	route_fd = (FILE *) 0;
    }
}

/********************************************************************
 *
 * open_route_table - open the interface to the route table
 */
static char route_delims[] = " \t\n";

static int open_route_table (void)
{
    char *path;

    close_route_table();

    path = "/proc/net/route";
    route_fd = fopen (path, "r");
    if (route_fd == NULL) {
	perror("can't open routing table %s");
	return 0;
    }

    route_dev_col = 0;		/* default to usual columns */
    route_dest_col = 1;
    route_gw_col = 2;
    route_flags_col = 3;
    route_mask_col = 7;
    route_num_cols = 8;

    /* parse header line */
    if (fgets(route_buffer, sizeof(route_buffer), route_fd) != 0) {
	char *p = route_buffer, *q;
	int col;
	for (col = 0; col < ROUTE_MAX_COLS; ++col) {
	    int used = 1;
	    if ((q = strtok(p, route_delims)) == 0)
		break;
	    if (strcasecmp(q, "iface") == 0)
		route_dev_col = col;
	    else if (strcasecmp(q, "destination") == 0)
		route_dest_col = col;
	    else if (strcasecmp(q, "gateway") == 0)
		route_gw_col = col;
	    else if (strcasecmp(q, "flags") == 0)
		route_flags_col = col;
	    else if (strcasecmp(q, "mask") == 0)
		route_mask_col = col;
	    else
		used = 0;
	    if (used && col >= route_num_cols)
		route_num_cols = col + 1;
	    p = NULL;
	}
    }

    return 1;
}

/********************************************************************
 *
 * read_route_table - read the next entry from the route table
 */

static int read_route_table(struct rtentry *rt)
{
    char *cols[ROUTE_MAX_COLS], *p;
    int col;

    memset (rt, '\0', sizeof (struct rtentry));

    if (fgets (route_buffer, sizeof (route_buffer), route_fd) == (char *) 0)
	return 0;

    p = route_buffer;
    for (col = 0; col < route_num_cols; ++col) {
	cols[col] = strtok(p, route_delims);
	if (cols[col] == NULL)
	    return 0;		/* didn't get enough columns */
	p = NULL;
    }

    SIN_ADDR(rt->rt_dst) = strtoul(cols[route_dest_col], NULL, 16);
    SIN_ADDR(rt->rt_gateway) = strtoul(cols[route_gw_col], NULL, 16);
    SIN_ADDR(rt->rt_genmask) = strtoul(cols[route_mask_col], NULL, 16);

    rt->rt_flags = (short) strtoul(cols[route_flags_col], NULL, 16);
    rt->rt_dev   = cols[route_dev_col];

    return 1;
}

static void set_sockaddr(struct sockaddr_in *sin, unsigned long addr, unsigned short port)
{
	sin->sin_family = AF_INET;
	sin->sin_addr.s_addr = addr;
	sin->sin_port = port;
}

static int changeRoute(unsigned long ulIP,
			unsigned long ulGateway,
			unsigned long ulNetmask,
			short sMetric,
			const char *pcInterface,
			int iAction)
{
	struct rtentry rt;
	int fd;

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return FALSE;

	memset(&rt, 0, sizeof(rt));
	set_sockaddr((struct sockaddr_in *) &rt.rt_dst, ulIP, 0);
	set_sockaddr((struct sockaddr_in *) &rt.rt_gateway, ulGateway, 0);
	set_sockaddr((struct sockaddr_in *) &rt.rt_genmask, ulNetmask, 0);
	rt.rt_metric = sMetric + 1; 
	if( pcInterface && pcInterface[0] )
		rt.rt_dev = (char *)pcInterface;

	rt.rt_flags = RTF_UP;
	if (ulGateway != 0L)
    		rt.rt_flags |= RTF_GATEWAY;
	if ((ulIP & 0xFF000000) != 0L && ulNetmask == 0xFFFFFFFF)
		rt.rt_flags |= RTF_HOST;

	if (ioctl(fd, iAction, &rt) == -1)
	{
		fprintf(stderr, "change route (dst=%08x gw=%08x mask=%08x metric=%d if=%s action=%d) ioctl errno = %d\n", 
			ulIP, ulGateway, ulNetmask, sMetric, pcInterface, iAction, errno);
		close(fd);
		return FALSE;
	}
	close(fd);
	return TRUE;
}

int addRoute(unsigned long ulIP,
				unsigned long ulGateway,
				unsigned long ulNetmask,
				short sMetric,
				const char *pcInterface)
{
	return changeRoute(ulIP, ulGateway, ulNetmask, sMetric, pcInterface, SIOCADDRT);
}

int delRoute(unsigned long ulIP,
				unsigned long ulGateway,
				unsigned long ulNetmask,
				short sMetric,
				const char *pcInterface)
{
	return changeRoute(ulIP, ulGateway, ulNetmask, sMetric, pcInterface, SIOCDELRT);
}

/********************************************************************
 *
 * defaultroute_exists - determine if there is a default route
 */

static int defaultroute_exists (struct rtentry *rt)
{
    int result = 0;

    if (!open_route_table())
	return 0;

    while (read_route_table(rt) != 0) {
	if ((rt->rt_flags & RTF_UP) == 0)
	    continue;

	if( SIN_ADDR(rt->rt_genmask) != 0)
	    continue;
	if (SIN_ADDR(rt->rt_dst) == 0L) {
	    result = 1;
	    break;
	}
    }

    close_route_table();

    return result;
}


/*
 * have_route_to - determine if the system has any route to
 * a given IP address.  ddr' is in network byte order.
 * Return value is 1 if yes, 0 if no, -1 if don't know.
 * For demand mode to work properly, we have to ignore routes
 * through our own interface.
 */
int have_route_to(u_int32_t addr)
{
    struct rtentry rt;
    int result = 0;

    if (!open_route_table())
	return -1;		/* don't know */

    while (read_route_table(&rt)) {
	if ((rt.rt_flags & RTF_UP) == 0 )
	    continue;
	if ((addr & SIN_ADDR(rt.rt_genmask)) == SIN_ADDR(rt.rt_dst)) {
	    result = 1;
	    break;
	}
    }

    close_route_table();
    return result;
}

int getDefaultGateway (void)
{
	struct rtentry rt;
	
	if( defaultroute_exists(&rt) )
	{
		struct in_addr old_gateway =
	          ((struct sockaddr_in *) (&rt.rt_gateway))-> sin_addr;
	          
	        return old_gateway.s_addr;
	}
	
	return 0;
}

/********************************************************************
 *
 * sifdefaultroute - assign a default route through the address given.
 */
//int req_unit = 1;

int setDefaultGateway (char *ifname, u_int32_t gateway, u_int32_t metric)
{
    struct rtentry rt;
    int sockfd;

    sockfd = socket(AF_INET, SOCK_DGRAM, 0 );
    if( sockfd< 0 )
    {
    	printf("Create socket failed\n");
    	close(sockfd);
    	return -1;
    }
    //printf("\nsetDefaultGateway if=%s gw=%08x metric=%d \n", ifname, gateway, metric);
    memset (&rt, '\0', sizeof (rt));
    SET_SA_FAMILY (rt.rt_dst,     AF_INET);
    SET_SA_FAMILY (rt.rt_gateway, AF_INET);

    if( ifname )
    	rt.rt_dev = ifname;
    rt.rt_metric = (metric)+1; /* norton */
    
    SET_SA_FAMILY (rt.rt_genmask, AF_INET);
    SIN_ADDR(rt.rt_genmask) = 0L;
    SIN_ADDR(rt.rt_gateway) = gateway;

    rt.rt_flags = RTF_UP | RTF_GATEWAY;
    if (ioctl(sockfd, SIOCADDRT, &rt) < 0) {
	perror("default route ioctl(SIOCADDRT)");
	close(sockfd);
	return -1;
    }

    
    return 0;
}

/********************************************************************
 *
 * cifdefaultroute - delete a default route through the address given.
 */

int clearDefaultGateway (char *ifname, u_int32_t gateway)
{
    struct rtentry rt;
    int sockfd;
    
    sockfd = socket(AF_INET, SOCK_DGRAM, 0 );
    if( sockfd< 0 )
    {
    	printf("Create socket failed\n");
    	close(sockfd);
    	return -1;
    }
    //printf("\nclearDefaultGateway if=%s gw=%08x\n", ifname, gateway);

    memset (&rt, '\0', sizeof (rt));
    SET_SA_FAMILY (rt.rt_dst,     AF_INET);
    SET_SA_FAMILY (rt.rt_gateway, AF_INET);

    if( ifname )
    	rt.rt_dev = ifname; /* norton , exactly match interface, especial in multiple pppoe */

    SET_SA_FAMILY (rt.rt_genmask, AF_INET);
    SIN_ADDR(rt.rt_genmask) = 0L;
    SIN_ADDR(rt.rt_gateway) = gateway;
    rt.rt_flags = RTF_UP | RTF_GATEWAY;

    if (ioctl(sockfd, SIOCDELRT, &rt) < 0 && errno != ESRCH) {
	perror("default route ioctl(SIOCDELRT):");
	close(sockfd);
	return 0;
    }
    return 1;
}

